@@ -4,14 +4,10 @@ import (
 	"io"
 	"net"
 	"strings"
+
+	"github.com/go-redis/redis/internal/proto"
 )
 
-const Nil = RedisError("redis: nil")
-
-type RedisError string
-
-func (e RedisError) Error() string { return string(e) }
-
 func IsRetryableError(err error, retryNetError bool) bool {
 	if IsNetworkError(err) {
 		return retryNetError
@@ -30,7 +26,7 @@ func IsRetryableError(err error, retryNetError bool) bool {
 }
 
 func IsRedisError(err error) bool {
-	_, ok := err.(RedisError)
+	_, ok := err.(proto.RedisError)
 	return ok
 }
 
@@ -42,6 +38,10 @@ func IsNetworkError(err error) bool {
 	return ok
 }
 
+func IsReadOnlyError(err error) bool {
+	return strings.HasPrefix(err.Error(), "READONLY ")
+}
+
 func IsBadConn(err error, allowTimeout bool) bool {
 	if err == nil {
 		return false
